(* Abstract Syntax Tree and functions for printing it *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or | Mod | DAdd | DSub | DMult | DDiv | DMod | In

(* Negation equals = Notequals | Negation op = not op *)
type uop = Neg | Not | Incr | Decr

(* define the range of types available *)
type typ =
    Int
  | Double
  | Bool
  | Char
  | String
  | StructType of string
  | Map of typ * typ
  | Graph of typ * typ
  | Edge of typ * typ
  | Array of typ

type bind = typ * string

type struct_field = string * typ
type struct_decl = string * struct_field list

type expr =
    IntLit of int
  | DubLit of float
  | True
  | False
  | CharLit of string
  | StrLit of string
  | StructLit of (string * expr) list
  | MapLit of (expr * expr) list
  | GraphLit of expr list
  | EdgeLit of expr * expr * expr
  | ArrayLit of expr list
  | Id of string
  | Index of string * expr
  | StructFieldAccess of string * string
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Assign of expr * expr
  | Reassign of expr * expr
  | Call of string * expr list
  | Noexpr

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt list * stmt list
  | IfElseifs of expr * stmt list * (expr * stmt list) list * stmt list
  | For of expr * expr * expr * stmt list
  | While of expr * stmt list
  | Continue
  | Break

type func_decl = {
    typ : typ;
    fname : string;
    formals : bind list;
    locals : bind list;
    body : stmt list;
  }

(* vdecl_list, sdecl_list, fdecl_list *)
type program = bind list * struct_decl list * func_decl list

(* Pretty-printing functions *)

let string_of_op = function
    Add -> "+"
  | DAdd -> "+."
  | Sub -> "-"
  | DSub -> "-."
  | Mult -> "*"
  | DMult -> "*."
  | Div -> "/"
  | DDiv -> "/."
  | Mod -> "%"
  | DMod -> "%."
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"
  | In -> "in"


let string_of_uop = function
    Neg -> "-"
  | Not -> "!"
  | Incr -> "++"
  | Decr -> "--"

let rec string_of_expr = function
    IntLit(l) -> string_of_int l
  | DubLit(d) -> string_of_float d
  | True -> "true"
  | False -> "false"
  | CharLit(c) -> c
  | StrLit(s) -> s
  | ArrayLit(l) -> "[" ^ String.concat "," (List.map string_of_expr (List.rev l)) ^ "]"
  | StructLit(l) -> "{" ^ String.concat "," (List.map (fun (f, v) -> f ^ ":" ^ (string_of_expr v)) (List.rev l)) ^ "}"
  | MapLit(l) -> "[" ^ String.concat "," (List.map (fun (k, v) -> (string_of_expr k) ^ ":" ^ (string_of_expr v)) (List.rev l)) ^ "]"
  | GraphLit(l) -> "{" ^ String.concat ", " (List.map string_of_expr (List.rev l)) ^ "}"
  | EdgeLit(a, b, c) -> "(" ^ (string_of_expr a) ^ ", " ^ (string_of_expr b) ^ ", " ^ (string_of_expr c) ^ ")"
  | Id(s) -> s
  | Index(s, e) -> s ^ "[" ^ (string_of_expr e) ^ "]"
  | StructFieldAccess (s, t) -> s ^ "." ^ t
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Assign(l, r) -> string_of_expr l ^ " := " ^ string_of_expr r
  | Reassign(l, r) -> string_of_expr l ^ " = " ^ string_of_expr r
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt (List.rev stmts)) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, [Block([])]) -> "if (" ^ string_of_expr e ^ ")\n  " ^
      String.concat "  " (List.map string_of_stmt (List.rev s))
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n  " ^
      String.concat "  " (List.map string_of_stmt (List.rev s1)) ^
      "else\n  " ^ String.concat "  " (List.map string_of_stmt (List.rev s2))
  | IfElseifs(e1, s1, l, [Block([])]) -> "if (" ^ string_of_expr e1 ^ ")\n  " ^
      String.concat "  " (List.map string_of_stmt (List.rev s1)) ^ String.concat ""
      (List.map (fun (e2, s2) -> ("elseif (" ^ string_of_expr e2 ^ ")\n  " ^ String.concat "  " (List.map string_of_stmt (List.rev s2)))) l)
  | IfElseifs(e1, s1, l, s3) ->  "if (" ^ string_of_expr e1 ^ ")\n  " ^
      String.concat "  " (List.map string_of_stmt (List.rev s1)) ^ String.concat ""
      (List.map (fun (e2, s2) -> ("elseif (" ^ string_of_expr e2 ^ ")\n  " ^ String.concat "  " (List.map string_of_stmt (List.rev s2)))) l) ^
      "else\n  " ^ String.concat "  " (List.map string_of_stmt (List.rev s3))
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ String.concat "" (List.map string_of_stmt (List.rev s))
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ String.concat "" (List.map string_of_stmt (List.rev s))
  | Continue -> "continue;"
  | Break -> "break;"

let rec string_of_typ = function
    Int -> "int"
  | Bool -> "bool"
  | String -> "str"
  | Double -> "double"
  | Char -> "char"
  | Array(t) -> (string_of_typ t) ^ "[]"
  | StructType(t) -> t
  | Map(a, b) -> "map<" ^ (string_of_typ a) ^ ", " ^ (string_of_typ b) ^ ">"
  | Graph(a, b) -> "graph<" ^ (string_of_typ a) ^ ", " ^ (string_of_typ b) ^ ">"
  | Edge(a, b) -> "edge<" ^ (string_of_typ a) ^ ", " ^ (string_of_typ b) ^ ">"

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.locals) ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_sdecl (struct_name, fields) =
  struct_name ^ "{" ^ String.concat "," (List.map (fun (f, t) -> f ^
      ":" ^ (string_of_typ t)) fields)  ^ "}"

let string_of_program (vars, structs, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_sdecl structs) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)

(* Helper functions *)
let fst_of_3 (a,_,_) = a
let snd_of_3 (_,a,_) = a
let thd_of_3 (_,_,a) = a
