(* Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast
open Sast

module StringMap = Map.Make(String)

(* translate : Sast.program -> Llvm.module *)
let translate (globals, _, functions) =
  let context    = L.global_context () in
  let llmem = L.MemoryBuffer.of_file "gae.bc" in
  let llm = Llvm_bitreader.parse_bitcode context llmem in
  (* Create the LLVM compilation module into which
     we will generate code *)
  let the_module = L.create_module context "gae" in

  (* Get types from the context *)
  let i32_t      = L.i32_type    context
  and i8_t       = L.i8_type     context
  and i1_t       = L.i1_type     context
  and float_t    = L.double_type context
  and str_t      = L.pointer_type (L.i8_type context)
  and void_ptr_t = L.pointer_type (L.i8_type context)
  and arr_t      = L.pointer_type (match L.type_by_name llm "struct.Array" with
      None -> raise (Failure "struct.Array isn't defined.")
    | Some x -> x)
  and arr_elmt_t = L.pointer_type (match L.type_by_name llm "struct.Array_element" with
      None -> raise (Failure "struct.Array_element isn't defined.")
    | Some x -> x)
  and map_t      = L.pointer_type (match L.type_by_name llm "struct.Map" with
      None -> raise (Failure "Missing implementation for struct Map")
    | Some t -> t)
  and struct_t   = L.pointer_type (match L.type_by_name llm "struct.Struct" with
      None -> raise (Failure "Struct implementation not defined")
    | Some t -> t)
  and struct_element_t = L.pointer_type (match L.type_by_name llm "struct.Struct_element" with
      None -> raise (Failure "Struct_element implementation not defined")
    | Some t -> t)
  and edge_t     = L.pointer_type (match L.type_by_name llm "struct.Edge" with
      None -> raise (Failure "Edge implementation not defined")
    | Some t -> t)
  and graph_t    = L.pointer_type (match L.type_by_name llm "struct.Graph" with
      None -> raise (Failure "Graph implementation not defined")
    | Some t -> t)
  in
  (* TODO: Add rest of tpyes *)

  (* Return the LLVM type for a gae type *)
  let ltype_of_typ = function
      A.Int       -> i32_t
    | A.Bool      -> i1_t
    | A.Double    -> float_t
    | A.String    -> str_t
    | A.Array _   -> arr_t
    | A.Map _ -> map_t
    | A.StructType _ -> struct_t
    | A.Edge (_, _) -> edge_t
    | A.Graph (_, _) -> graph_t
    | _ -> raise (Failure "not implemented")
  in
  (* TODO: add rest of types *)

  (* Create a map of global variables after creating each *)
  let global_vars : L.llvalue StringMap.t =
    let global_var m (t, n) =
      let init = match t with
          A.Double -> L.const_float (ltype_of_typ t) 0.0
        | _ -> L.const_int (ltype_of_typ t) 0
      in
      StringMap.add n (L.define_global n init the_module) m
    in
    List.fold_left global_var StringMap.empty globals
  in

  (* print functions *)
  let printf_t : L.lltype = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func : L.llvalue = L.declare_function "printf" printf_t the_module in

  (* misc function *)
  let int_of_double_t = L.function_type i32_t [| float_t |] in
  let int_of_double_f = L.declare_function "int_of_double" int_of_double_t the_module in
  let double_of_int_t = L.function_type float_t [| i32_t |] in
  let double_of_int_f = L.declare_function "double_of_int" double_of_int_t the_module in

  (* string functions *)
  let string_length_t = L.function_type i32_t [| str_t |] in
  let string_length_f = L.declare_function "string_length" string_length_t the_module in
  let string_concat_t = L.function_type str_t [| str_t; str_t |] in
  let string_concat_f = L.declare_function "string_concat" string_concat_t the_module in
  let string_equals_t = L.function_type i32_t [| str_t; str_t |] in
  let string_equals_f = L.declare_function "string_equals" string_equals_t the_module in

  (* array functions *)
  let arr_init_t = L.function_type arr_t [||] in
  let arr_init_f = L.declare_function "arr_init" arr_init_t the_module in
  let arr_set_contains_struct_t = L.function_type arr_t [| arr_t |] in
  let arr_set_contains_struct_f = L.declare_function "arr_set_contains_struct" arr_set_contains_struct_t the_module in
  let arr_append_t = L.function_type arr_elmt_t [| arr_t; void_ptr_t |] in
  let arr_append_f = L.declare_function "arr_append" arr_append_t the_module in
  let arr_get_t = L.function_type arr_elmt_t [| arr_t; i32_t |] in
  let arr_get_f = L.declare_function "arr_get" arr_get_t the_module in
  let arr_set_t = L.function_type arr_elmt_t [| arr_t; void_ptr_t; i32_t |] in
  let arr_set_f = L.declare_function "arr_set" arr_set_t the_module in
  let arr_length_t = L.function_type i32_t [| arr_t |] in
  let arr_length_f = L.declare_function "arr_length" arr_length_t the_module in
  let arr_contains_t = L.function_type i32_t [| arr_t; void_ptr_t |] in
  let arr_contains_f = L.declare_function "arr_contains" arr_contains_t the_module in

  (* map functions *)
  let map_init_t = L.function_type map_t [||] in
  let map_init_f = L.declare_function "map_init" map_init_t the_module in
  let map_set_contains_struct_t = L.function_type map_t [| map_t |] in
  let map_set_contains_struct_f = L.declare_function "map_set_contains_struct" map_set_contains_struct_t the_module in
  let map_get_t = L.function_type void_ptr_t [| map_t; void_ptr_t |] in
  let map_get_f = L.declare_function "map_get" map_get_t the_module in
  let map_set_t = L.function_type void_ptr_t [| map_t; void_ptr_t; void_ptr_t |] in
  let map_set_f = L.declare_function "map_set" map_set_t the_module in
  let map_length_t = L.function_type i32_t [| map_t |] in
  let map_length_f = L.declare_function "map_length" map_length_t the_module in
  let map_contains_t = L.function_type i32_t [| map_t; void_ptr_t |] in
  let map_contains_f = L.declare_function "map_contains" map_contains_t the_module in
  let map_get_keys_t = L.function_type arr_t [| map_t |] in
  let map_get_keys_f = L.declare_function "map_get_keys" map_get_keys_t the_module in

  (* struct functions *)
  let struct_init_t = L.function_type struct_t [||] in
  let struct_init_f = L.declare_function "struct_init" struct_init_t the_module in
  let struct_add_t = L.function_type struct_element_t [| struct_t; str_t; void_ptr_t |] in
  let struct_add_f = L.declare_function "struct_add" struct_add_t the_module in
  let struct_get_t = L.function_type void_ptr_t [| struct_t; str_t |] in
  let struct_get_f = L.declare_function "struct_get" struct_get_t the_module in
  let struct_set_t = L.function_type struct_element_t [| struct_t; str_t; void_ptr_t |] in
  let struct_set_f = L.declare_function "struct_set" struct_set_t the_module in
  let struct_equals_t = L.function_type i32_t [| struct_t; struct_t |] in
  let struct_equals_f = L.declare_function "struct_equals" struct_equals_t the_module in

  (* edge functions *)
  let edge_init_t = L.function_type edge_t [| struct_t; struct_t; struct_t |] in
  let edge_init_f = L.declare_function "edge_init" edge_init_t the_module in
  let edge_get_src_t = L.function_type struct_t [| edge_t |] in
  let edge_get_src_f = L.declare_function "edge_get_src" edge_get_src_t the_module in
  let edge_get_dst_t = L.function_type struct_t [| edge_t |] in
  let edge_get_dst_f = L.declare_function "edge_get_dst" edge_get_dst_t the_module in
  let edge_get_val_t = L.function_type struct_t [| edge_t |] in
  let edge_get_val_f = L.declare_function "edge_get_val" edge_get_val_t the_module in
  let edge_set_src_t = L.function_type struct_t [| edge_t; struct_t |] in
  let edge_set_src_f = L.declare_function "edge_set_src" edge_set_src_t the_module in
  let edge_set_dst_t = L.function_type struct_t [| edge_t; struct_t |] in
  let edge_set_dst_f = L.declare_function "edge_set_dst" edge_set_dst_t the_module in
  let edge_set_val_t = L.function_type struct_t [| edge_t; struct_t |] in
  let edge_set_val_f = L.declare_function "edge_set_val" edge_set_val_t the_module in

  (* graph functions *)
  let graph_init_t = L.function_type graph_t [||] in
  let graph_init_f = L.declare_function "graph_init" graph_init_t the_module in
  let graph_add_edge_t = L.function_type edge_t [| graph_t; edge_t |] in
  let graph_add_edge_f = L.declare_function "graph_add_edge" graph_add_edge_t the_module in
  let graph_get_edges_t = L.function_type arr_t [| graph_t |] in
  let graph_get_edges_f = L.declare_function "graph_get_edges" graph_get_edges_t the_module in
  let graph_get_nodes_t = L.function_type arr_t [| graph_t |] in
  let graph_get_nodes_f = L.declare_function "graph_get_nodes" graph_get_nodes_t the_module in
  let graph_contains_node_t = L.function_type i32_t [| graph_t; struct_t |] in
  let graph_contains_node_f = L.declare_function "graph_contains_node" graph_contains_node_t the_module in

  (* Define each function (arguments and return type) so we can
     call it even before we've created its body *)
  let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
    let function_decl m fdecl =
      let name = fdecl.sfname
      and formal_types =
        Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.sformals)
      in
      let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m
    in
    List.fold_left function_decl StringMap.empty functions
  in

  (* Fill in the body of the given function *)
  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.sfname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in

    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and float_format_str = L.build_global_stringptr "%g\n" "fmt" builder
    and char_format_str = L.build_global_stringptr "%s\n" "fmt" builder
    and bool_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and string_format_str = L.build_global_stringptr "%s\n" "fmt" builder in

    (* Construct the function's "locals": formal arguments and locally
       declared variables.  Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals" map *)
    let local_vars =
      let add_formal m (t, n) p =
        L.set_value_name n p;
	      let local = L.build_alloca (ltype_of_typ t) n builder in
        ignore (L.build_store p local builder);
	      StringMap.add n local m

      (* Allocate space for any locally declared variables and add the
       * resulting registers to our map *)
      and add_local m (t, n) =
	      let local_var = L.build_alloca (ltype_of_typ t) n builder in
        StringMap.add n local_var m
      in

      let formals = List.fold_left2 add_formal StringMap.empty fdecl.sformals
          (Array.to_list (L.params the_function)) in
      List.fold_left add_local formals fdecl.slocals
    in

    (* Return the value for a variable or formal argument.
       Check local names first, then global names *)
    let lookup n = try StringMap.find n local_vars
                   with Not_found -> StringMap.find n global_vars
    in

    (* Construct code for an expression; return its value *)
    let rec expr builder ((styp, e) : sexpr) = match e with
	      SIntLit i   -> L.const_int i32_t i
      | SDubLit l   -> L.const_float float_t l
      | SStrLit s   -> L.build_global_stringptr s "str" builder
      | SCharLit c  -> L.build_global_stringptr c "char" builder
      | STrue       -> L.const_int i1_t 1
      | SFalse      -> L.const_int i1_t 0
      | SArrayLit contents -> let rec arr_fill arr = (function
              [] -> arr
            | sx :: rest ->
                let (typ, _) = sx in
                let data = (match typ with
                  | _ -> let data = L.build_malloc (ltype_of_typ typ) "data" builder in
                      let llvalue = expr builder sx in
                      ignore (L.build_store llvalue data builder); data)
                in
                let data = L.build_bitcast data void_ptr_t "data" builder in
                ignore (L.build_call arr_append_f [| arr; data |] "arr_append" builder);
                arr_fill arr rest)
            in
            let el_typ = fst (List.nth contents 0) in
            let arr = match el_typ with
                A.StructType(_) ->
                  let arr = L.build_call arr_init_f [||] "arr_init" builder in
                  L.build_call arr_set_contains_struct_f [| arr |] "arr_set_contains_struct" builder
              | _ ->
                  L.build_call arr_init_f [||] "arr_init" builder
            in
            ignore(arr_fill arr (List.rev contents)); arr
      | SNoexpr     -> L.const_int i32_t 0
      | SStructLit contents -> let rec struct_fill sstruct = (function
              [] -> sstruct
            | (sfield, svalue) :: rest ->
              let field = L.build_global_stringptr sfield "str" builder in
              let (vtyp, _) = svalue in
              let value = (match vtyp with
                | _ -> let value = L.build_malloc (ltype_of_typ vtyp) "value" builder in
                  let llvalue = expr builder svalue
                  in ignore (L.build_store llvalue value builder); value)
              in
              let value = L.build_bitcast value void_ptr_t "value" builder in
              ignore (L.build_call struct_add_f [| sstruct; field; value |] "struct_add" builder);
              struct_fill sstruct rest)
        in
        let sstruct = L.build_call struct_init_f [||] "struct_init" builder in
        ignore(struct_fill sstruct contents); sstruct
      | SMapLit contents ->
        let rec map_fill map_init = function
              [] -> map_init
            | (skey, svalue) :: rest ->
              let (ktyp, _) = skey in
              let key_ = (
                let key_ = L.build_malloc (ltype_of_typ ktyp) "key" builder in
                let llkey = expr builder skey in
                ignore (L.build_store llkey key_ builder); key_
              ) in
              let key = L.build_bitcast key_ void_ptr_t "key" builder in
              let (vtyp, _) = svalue in
              let value_ = (
                let data = L.build_malloc (ltype_of_typ vtyp) "val" builder in
                let llvalue = expr builder svalue in
                ignore (L.build_store llvalue data builder); data
              ) in
              let value = L.build_bitcast value_ void_ptr_t "val" builder in
              ignore (L.build_call map_set_f [| map_init; key; value |] "map_set" builder);
              map_fill map_init rest
        in
        let key_typ = fst(fst (List.nth contents 0)) in
        let map = match key_typ with
            A.StructType(_) ->
              let map = L.build_call map_init_f [||] "map_init" builder in
              L.build_call map_set_contains_struct_f [| map |] "map_set_contains_struct" builder
          | _ ->
              L.build_call map_init_f [||] "map_init" builder
        in
        ignore(map_fill map contents); map
      | SEdgeLit (src, dst, value) ->
        let src' = expr builder src
        and dst' = expr builder dst
        and val' = expr builder value in
          L.build_call edge_init_f [| src'; dst'; val' |] "edge_init" builder
      | SGraphLit contents ->
        let rec g_fill g = (function
          [] -> g
        | sx :: rest ->
            let edge = expr builder sx in
            ignore (L.build_call graph_add_edge_f [| g; edge |] "graph_add_edge" builder);
            g_fill g rest)
        in
        let g = L.build_call graph_init_f [||] "graph_init" builder in
        ignore(g_fill g (List.rev contents)); g
      | SId s       -> L.build_load (lookup s) s builder
      | SAssign ((_, le), (rt, re)) -> (match le with
            SId (s) ->
              let re' = expr builder (rt, re) in
              ignore(L.build_store re' (lookup s) builder); re'
          | SArrayIndex (id, e) ->
               let ltype = ltype_of_typ styp in
               let arr = L.build_load (lookup id) id builder in
               let index = expr builder e in
               let data = L.build_malloc ltype "data" builder in
               ignore(L.build_store (expr builder (rt, re)) data builder);
               let data = L.build_bitcast data void_ptr_t "data" builder in
               ignore(L.build_call arr_set_f [| arr; data; index |] "arr_set" builder);
               data
          | SMapIndex (s, e) ->
              let map = L.build_load (lookup s) "map" builder in
              let (kt, _) = e in
              let key_ = (
                let key_ = L.build_malloc (ltype_of_typ kt) "key" builder in
                let llkey = expr builder e in
                ignore (L.build_store llkey key_ builder); key_
              ) in
              let key = L.build_bitcast key_ void_ptr_t "key" builder in
              let data_ = (
                let data = L.build_malloc (ltype_of_typ rt) "val" builder in
                let llvalue = expr builder (rt, re) in
                ignore (L.build_store llvalue data builder); data
              ) in
              let value = L.build_bitcast data_ void_ptr_t "val" builder in
              ignore (L.build_call map_set_f [| map; key; value |] "map_set" builder);
              value
          | SStructFieldAccess (s, t) ->
            let field = L.build_global_stringptr t "str" builder in
            let ltype = ltype_of_typ styp in
            let sstruct = L.build_load (lookup s) s builder in
            let value = L.build_malloc ltype "value" builder in
            ignore(L.build_store (expr builder (rt, re)) value builder);
            let value = L.build_bitcast value void_ptr_t "value" builder in
            ignore(L.build_call struct_set_f [| sstruct; field; value |] "struct_set" builder);
            value
          | _ -> raise (Failure "not implemented")
      )
      | SReassign (_, _) -> raise (Failure "Reassign should never be called in codegen")
      | SStructFieldAccess (s, t) ->
          let field = L.build_global_stringptr t "str" builder in
          let ltype = ltype_of_typ styp in
          let sstruct = L.build_load (lookup s) s builder in
          let data_ptr = L.build_call struct_get_f [| sstruct; field |] "struct_get" builder in
          let data_ptr = L.build_bitcast data_ptr (L.pointer_type ltype) "value" builder in
          L.build_load data_ptr "value" builder
      | SMapIndex (s, e) -> (match e with
          (_ as kt, _) ->
            let ltype = ltype_of_typ kt in
            let key_ = (
              let key_ = L.build_malloc ltype "key" builder in
              let llkey = expr builder e in
              ignore (L.build_store llkey key_ builder); key_
            ) in
            let key = L.build_bitcast key_ void_ptr_t "key" builder in
            let map = L.build_load (lookup s) "map" builder in
            let ptr = L.build_call map_get_f [| map; key |] "map_get" builder in
            let ptr = L.build_bitcast ptr (L.pointer_type (ltype_of_typ styp)) "data" builder in
            L.build_load ptr "data" builder
        )
      | SArrayIndex (id, e) ->
            let ltype = ltype_of_typ styp in
            let arr = L.build_load (lookup id) id builder in
            let index = expr builder e in
            let data_ptr = L.build_call arr_get_f [| arr; index |] "arr_get" builder in
            let data_ptr = L.build_bitcast data_ptr (L.pointer_type ltype) "data" builder in
            L.build_load data_ptr "data" builder
      | SBinop ((A.Double,_ ) as e1, op, e2) ->
      	  let e1' = expr builder e1
      	  and e2' = expr builder e2 in
      	  (match op with
      	    A.DAdd     -> L.build_fadd
      	  | A.DSub     -> L.build_fsub
      	  | A.DMult    -> L.build_fmul
      	  | A.DDiv     -> L.build_fdiv
      	  | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
      	  | A.Neq     -> L.build_fcmp L.Fcmp.One
      	  | A.Less    -> L.build_fcmp L.Fcmp.Olt
      	  | A.Leq     -> L.build_fcmp L.Fcmp.Ole
      	  | A.Greater -> L.build_fcmp L.Fcmp.Ogt
      	  | A.Geq     -> L.build_fcmp L.Fcmp.Oge
      	  | A.And | A.Or ->
      	      raise (Failure "internal error: semant should have rejected and/or on float")
          | _ -> raise (Failure "not implemented")
      	  ) e1' e2' "tmp" builder
      | SBinop ((A.String, _) as e1, op, e2) when op = A.Add || op = A.Equal || op = A.Neq ->
          let e1' = expr builder e1
          and e2' = expr builder e2 in
          (match op with
            A.Add     -> L.build_call string_concat_f [| e1'; e2' |] "string_concat" builder
          | A.Equal   -> (L.build_icmp L.Icmp.Eq) (L.const_int i32_t 0)
                (L.build_call string_equals_f [| e1'; e2' |] "string_equals" builder) "tmp" builder
          | A.Neq     -> (L.build_icmp L.Icmp.Ne) (L.const_int i32_t 0)
                (L.build_call string_equals_f [| e1'; e2' |] "string_equals" builder) "tmp" builder
          | _ -> raise (Failure "not implemented")
          )
      | SBinop ((t1, e1), op, (t2, e2)) when
        (op = A.Equal || op = A.Neq) &&
        (match (t1, t2) with
            (A.StructType(t1'), A.StructType(t2')) -> t1' = t2'
          | _ -> false
        ) ->
          let e1' = expr builder (t1, e1)
          and e2' = expr builder (t2, e2) in
          (match op with
            A.Equal   -> (L.build_icmp L.Icmp.Eq) (L.const_int i32_t 1)
                (L.build_call struct_equals_f [| e1'; e2' |] "struct_equals" builder) "tmp" builder
          | A.Neq     -> (L.build_icmp L.Icmp.Ne) (L.const_int i32_t 1)
                (L.build_call struct_equals_f [| e1'; e2' |] "struct_equals" builder) "tmp" builder
          | _ -> raise (Failure "not implemented")
          )
      | SBinop ((t1, e1), op, (t2, e2)) when op = A.In ->
        let e1' = expr builder (t1, e1)
        and e2' = expr builder (t2, e2) in
        let t1' = ltype_of_typ t1 in
        let val_ = L.build_malloc t1' "val" builder in
        ignore (L.build_store e1' val_ builder);
        let value = L.build_bitcast val_ void_ptr_t "val" builder in
        (match t2 with
            A.Array(_) ->
              L.build_call arr_contains_f [| e2'; value |] "arr_contains" builder
          | A.Map(_, _) ->
              L.build_call map_contains_f [| e2'; value |] "map_contains" builder
          | A.Graph(_, _) ->
              (match t1 with
                  A.StructType(_) ->
                    let value = L.build_bitcast value struct_t "struct" builder in
                    (L.build_icmp L.Icmp.Ne) (L.const_int i32_t 1) (L.build_call graph_contains_node_f [| e2'; value |] "graph_contains_node" builder) "tmp" builder
                (* | Edge(_, _) -> *)
                | _ -> raise (Failure ("expected edge or node, received " ^ A.string_of_typ t1))
              )
          | _ -> raise (Failure "not implemented")
        )
      | SBinop ((A.Int, _) as e1, op, e2) ->
      	  let e1' = expr builder e1
      	  and e2' = expr builder e2 in
      	  (match op with
      	    A.Add     -> L.build_add
      	  | A.Sub     -> L.build_sub
      	  | A.Mult    -> L.build_mul
          | A.Div     -> L.build_sdiv
          | A.Mod     -> L.build_srem
      	  | A.And     -> L.build_and
      	  | A.Or      -> L.build_or
      	  | A.Equal   -> L.build_icmp L.Icmp.Eq
      	  | A.Neq     -> L.build_icmp L.Icmp.Ne
      	  | A.Less    -> L.build_icmp L.Icmp.Slt
      	  | A.Leq     -> L.build_icmp L.Icmp.Sle
      	  | A.Greater -> L.build_icmp L.Icmp.Sgt
      	  | A.Geq     -> L.build_icmp L.Icmp.Sge
          | _ -> raise (Failure "not implemented")
      	  ) e1' e2' "tmp" builder
      | SBinop ((A.Bool, _) as e1, op, e2) ->
      	  let e1' = expr builder e1
      	  and e2' = expr builder e2 in
      	  (match op with
            A.And -> L.build_and
          | A.Or -> L.build_or
          | _ -> raise (Failure "not implemented")
          ) e1' e2' "tmp" builder
      | SBinop (_,_,_) as e -> raise (Failure ("binop not implemented for this expr " ^ string_of_sexpr (styp, e)))
      | SUnop(op, ((t, _) as e)) ->
          let e' = expr builder e in
	        (match op with
	          A.Neg when t = A.Double -> L.build_fneg e' "tmp" builder
	        | A.Neg                   -> L.build_neg e' "tmp" builder
          | A.Not                   -> L.build_not e' "tmp" builder
          | A.Incr -> raise (Failure "Incr should never be called in codegen")
          | A.Decr -> raise (Failure "Decr should never be called in codegen")
          )
      | SCall ("printi", [e]) -> L.build_call printf_func [| int_format_str ; (expr builder e) |] "printf" builder
      | SCall ("printb", [e]) -> L.build_call printf_func [| bool_format_str ; (expr builder e) |] "printf" builder
      | SCall ("prints", [e]) -> L.build_call printf_func [| string_format_str ; (expr builder e) |] "printf" builder
      | SCall ("printc", [e]) -> L.build_call printf_func [| char_format_str ; (expr builder e) |] "printf" builder
      | SCall ("printd", [e]) -> L.build_call printf_func [| float_format_str ; (expr builder e) |] "printf" builder
      | SCall ("lena", [a]) -> L.build_call arr_length_f [| expr builder a |] "lena" builder
      | SCall ("lenm", [m]) -> L.build_call map_length_f [| expr builder m |] "lenm" builder
      | SCall ("lens", [s]) -> L.build_call string_length_f [| expr builder s |] "lens" builder
      | SCall ("getKeys", [m]) -> L.build_call map_get_keys_f [| expr builder m |] "map_get_keys" builder
      | SCall ("map_init", []) ->
        (* L.build_call map_init_f [||] "map_init" builder *)
        (match styp with
            A.Map((_ as s), _) when (match s with A.StructType(_) -> true | _ -> false)->
              let map = L.build_call map_init_f [||] "map_init" builder in
              L.build_call map_set_contains_struct_f [| map |] "map_set_contains_struct" builder
          | _ -> L.build_call map_init_f [||] "map_init" builder
        )
      | SCall ("graph_init", []) -> L.build_call graph_init_f [||] "graph_init" builder
      | SCall ("arr_init", []) ->
        (* L.build_call arr_init_f [||] "arr_init" builder *)
        (match styp with
            A.Array(_ as s) when (match s with A.StructType(_) -> true | _ -> false)->
              let arr = L.build_call arr_init_f [||] "arr_init" builder in
              L.build_call arr_set_contains_struct_f [| arr |] "arr_set_contains_struct" builder
          | _ -> L.build_call arr_init_f [||] "arr_init" builder
        )
      | SCall ("getSrc", [e]) -> L.build_call edge_get_src_f [| expr builder e |] "edge_get_src" builder
      | SCall ("getDst", [e]) -> L.build_call edge_get_dst_f [| expr builder e |] "edge_get_dst" builder
      | SCall ("getVal", [e]) -> L.build_call edge_get_val_f [| expr builder e |] "edge_get_val" builder
      | SCall ("setSrc", [e; n]) -> L.build_call edge_set_src_f [| expr builder e; expr builder n |] "edge_set_src" builder
      | SCall ("setDst", [e; n]) -> L.build_call edge_set_dst_f [| expr builder e; expr builder n |] "edge_set_dst" builder
      | SCall ("setVal", [e; v]) -> L.build_call edge_set_val_f [| expr builder e; expr builder v |] "edge_set_val" builder
      | SCall ("append", [a; e]) ->
        let (typ, _) = e in
        let data = L.build_malloc (ltype_of_typ typ) "data" builder in
        ignore (L.build_store (expr builder e) data builder);
        let data = L.build_bitcast data void_ptr_t "data" builder in
        L.build_call arr_append_f [| expr builder a; data |] "arr_append" builder
      | SCall ("addEdge", [g; e]) -> L.build_call graph_add_edge_f [| expr builder g; expr builder e |] "graph_add_edge" builder
      | SCall ("getEdges", [g]) -> L.build_call graph_get_edges_f [| expr builder g |] "graph_get_edges" builder
      | SCall ("getNodes", [g]) -> L.build_call graph_get_nodes_f [| expr builder g |] "graph_get_nodes" builder
      | SCall ("int_of_double", [e]) -> L.build_call int_of_double_f [| expr builder e |] "int_of_double" builder
      | SCall ("double_of_int", [e]) -> L.build_call double_of_int_f [| expr builder e |] "double_of_int" builder
      | SCall (f, args) ->
        let (fdef, fdecl) = StringMap.find f function_decls in
        let llargs = List.rev (List.map (expr builder) (List.rev args)) in
	      let result = (match fdecl.styp with
            | _ -> f ^ "_result") in
        L.build_call fdef (Array.of_list llargs) result builder
    in

    (* LLVM insists each basic block end with exactly one "terminator"
       instruction that transfers control.  This function runs "instr builder"
       if the current block does not already have a terminator.  Used,
       e.g., to handle the "fall off the end of the function" case. *)
    let add_terminal builder instr =
      match L.block_terminator (L.insertion_block builder) with
	      Some _ -> ()
      | None -> ignore (instr builder) in

    (* Build the code for the given statement; return the builder for
       the statement's successor (i.e., the next instruction will be built
       after the one generated by this call) *)

    let rec stmt builder = function
        SBlock sl -> List.fold_left stmt builder sl
      | SExpr e -> ignore(expr builder e); builder
      | SReturn e -> ignore(L.build_ret (expr builder e) builder );
                     builder
      | SIf (predicate, then_block, else_block) ->
         let bool_val = expr builder predicate in
	       let merge_bb = L.append_block context "merge" the_function in
         let build_br_merge = L.build_br merge_bb in (* partial function *)

         let then_bb = L.append_block context "then" the_function in
         let then_builder = (List.map (fun (then_stmt) -> stmt (L.builder_at_end context then_bb) then_stmt) (List.rev then_block)) in
         ignore((List.map (fun (then_stmt) -> add_terminal then_stmt build_br_merge) then_builder));

      	 let else_bb = L.append_block context "else" the_function in
         let else_builder = (List.map (fun (else_stmt) -> stmt (L.builder_at_end context else_bb) else_stmt) (List.rev else_block)) in
         ignore((List.map (fun (else_stmt) -> add_terminal else_stmt build_br_merge) else_builder));

      	 ignore(L.build_cond_br bool_val then_bb else_bb builder);
      	 L.builder_at_end context merge_bb
      | SIfElseifs _ -> raise (Failure "elseifs not defined.")
      | SBreak -> raise (Failure "break not defined")
      | SContinue -> raise (Failure "continue not defined")
      | SWhile (predicate, body) ->
    	   let pred_bb = L.append_block context "while" the_function in
    	   ignore(L.build_br pred_bb builder);

    	   let body_bb = L.append_block context "while_body" the_function in
         let while_builder = (List.map (fun (while_stmt) -> stmt (L.builder_at_end context body_bb) while_stmt) (List.rev body)) in
         ignore((List.map (fun (while_stmt) -> add_terminal while_stmt (L.build_br pred_bb)) while_builder));

    	   let pred_builder = L.builder_at_end context pred_bb in
    	   let bool_val = expr pred_builder predicate in

    	   let merge_bb = L.append_block context "merge" the_function in
    	   ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
    	   L.builder_at_end context merge_bb
      | SFor (e1, e2, e3, body) -> stmt builder
	       ( SBlock [SExpr e1 ; SWhile (e2, [SBlock ((List.rev body) @ [SExpr e3])]) ] )
    in

    (* Build the code for each statement in the function *)
    let builder = stmt builder (SBlock fdecl.sbody) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.styp with
       (* A.Void -> L.build_ret_void *)
      | A.Double -> L.build_ret (L.const_float float_t 0.0)
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body functions;
  the_module
