/* Ocamlyacc parser for MicroC */

%{
open Ast
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA PERIOD
%token PLUS MINUS TIMES DIVIDE ASSIGN NOT
%token EQ NEQ LT LEQ GT GEQ AND OR
%token RETURN IF ELSEIF ELSE FOR WHILE
%token INT BOOL DOUBLE CHAR STRING STRUCT MAP GRAPH EDGE
%token MOD DPLUS DMINUS DTIMES DDIVIDE DMOD IN
%token REASSIGN FUNC COLON INCR DECR CONTINUE
%token BREAK LBRACKET RBRACKET TRUE FALSE
%token <int> INTLIT
%token <float> DUBLIT
%token <string> CHARLIT
%token <string> STRLIT
%token <string> ID
%token <string> STRUCTID
%token <string> FIELD
%token EOF

%right ASSIGN REASSIGN
%left IN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS DPLUS DMINUS
%left TIMES DIVIDE MOD DTIMES DDIVIDE DMOD
%right NOT NEG

%start program
%type <Ast.program> program

%%

program:
  decls EOF { $1 }

decls:
   /* nothing */ { [], [], [] }
 | decls vdecl { ($2 :: fst_of_3 $1), snd_of_3 $1, thd_of_3 $1 }
 | decls sdecl { fst_of_3 $1, ($2 :: snd_of_3 $1), thd_of_3 $1 }
 | decls fdecl { fst_of_3 $1, snd_of_3 $1, ($2 :: thd_of_3 $1) }

fdecl:
  FUNC ID LPAREN formals_opt RPAREN typ LBRACE vdecl_list stmt_list RBRACE
     { { typ = $6;
	 fname = $2;
	 formals = $4;
	 locals = List.rev $8;
	 body = List.rev $9 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    typ ID                   { [($1,$2)] }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

vdecl_list:
    /* nothing */     { [] }
  | vdecl_list vdecl  { $2 :: $1 }

typ:
    base_typ_or_struct { $1 }
  | MAP LT base_typ_or_struct COMMA base_typ_or_struct GT { Map($3, $5) } /* map<int, int> */
  | GRAPH LT STRUCTID COMMA STRUCTID GT { Graph(StructType($3), StructType($5)) } /* graph<Int, Int> */
  | base_typ_or_struct LBRACKET RBRACKET { Array($1) } /* int[] */

base_typ: /* non-container types only */
    INT { Int }
  | DOUBLE { Double }
  | BOOL { Bool }
  | CHAR { Char }
  | STRING { String }

base_typ_or_struct:
    base_typ { $1 }
  | STRUCTID { StructType($1) }
  | EDGE LT STRUCTID COMMA STRUCTID GT { Edge(StructType($3), StructType($5)) }

vdecl:
    typ ID SEMI { ($1, $2) }

sdecl:
    STRUCT STRUCTID LBRACE struct_field_list RBRACE { ($2, $4) } /*(struct_name, list of field:type in the struct)*/

struct_field_list:
    struct_field                            { [$1] }
  | struct_field_list COMMA struct_field    { $3 :: $1 }

struct_field:
    FIELD base_typ { ($1, $2) } /* returns (field_name, field_type) */

stmt_list:
    stmt            { [$1] }
  | stmt_list stmt  { $2 :: $1 }

loop_stmt_list:
    stmt                                { [$1] }
  | loop_stmt_list stmt                 { $2 :: $1 }
  | loop_stmt_list continue_or_break    { $2 :: $1 }

continue_or_break:
    CONTINUE SEMI   { Continue }
  | BREAK SEMI      { Break }

stmt:
    expr SEMI { Expr $1 }
  | RETURN SEMI { Return Noexpr }
  | RETURN expr SEMI { Return $2 }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF expr LBRACE stmt_list RBRACE { If($2, $4, [Block([])]) }
  | IF expr LBRACE stmt_list RBRACE ELSE LBRACE stmt_list RBRACE { If($2, $4, $8) }
  | IF expr LBRACE stmt_list RBRACE elseif_list { IfElseifs($2, $4, $6, [Block([])]) }
  | IF expr LBRACE stmt_list RBRACE elseif_list ELSE LBRACE stmt_list RBRACE { IfElseifs($2, $4, $6, $9)}
  | FOR expr SEMI expr SEMI expr LBRACE loop_stmt_list RBRACE
     { For($2, $4, $6, $8) }
  | WHILE expr LBRACE loop_stmt_list RBRACE { While($2, $4) }

elseif:
  ELSEIF expr LBRACE stmt_list RBRACE { ($2, $4) }

elseif_list:
    elseif { [$1] }
  | elseif_list elseif { $2 :: $1 }

expr:
    INTLIT              { IntLit($1) }
  | DUBLIT              { DubLit($1) }
  | CHARLIT             { CharLit($1) }
  | STRLIT              { StrLit($1) }
  | TRUE                { True }
  | FALSE               { False }
  | ID                  { Id($1) }
  | expr PLUS     expr  { Binop($1, Add,   $3) }
  | expr DPLUS    expr  { Binop($1, DAdd,  $3) }
  | expr MINUS    expr  { Binop($1, Sub,   $3) }
  | expr DMINUS   expr  { Binop($1, DSub,  $3) }
  | expr TIMES    expr  { Binop($1, Mult,  $3) }
  | expr DTIMES   expr  { Binop($1, DMult, $3) }
  | expr DIVIDE   expr  { Binop($1, Div,   $3) }
  | expr DDIVIDE  expr  { Binop($1, DDiv,  $3) }
  | expr MOD      expr  { Binop($1, Mod,   $3) }
  | expr DMOD     expr  { Binop($1, DMod,  $3) }
  | expr EQ       expr  { Binop($1, Equal, $3) }
  | expr NEQ      expr  { Binop($1, Neq,   $3) }
  | expr LT       expr  { Binop($1, Less,  $3) }
  | expr LEQ      expr  { Binop($1, Leq,   $3) }
  | expr GT       expr  { Binop($1, Greater, $3) }
  | expr GEQ      expr  { Binop($1, Geq,   $3) }
  | expr AND      expr  { Binop($1, And,   $3) }
  | expr OR       expr  { Binop($1, Or,    $3) }
  | expr IN       expr  { Binop($1, In,    $3) }
  | MINUS expr %prec NEG { Unop(Neg, $2) }
  | NOT expr         { Unop(Not, $2) }
  | ID INCR          { Assign(Id($1), (Binop(Id($1), Add, IntLit(1)))) }
  | ID DECR          { Assign(Id($1), (Binop(Id($1), Sub, IntLit(1)))) }
  | assignable ASSIGN expr   { Assign($1, $3) }
  | assignable REASSIGN expr   { Reassign($1, $3) }
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN { $2 }
  /* Struct literal */
  | struct_lit { $1 }
  /* Map literal */
  | LBRACKET kv_list RBRACKET { MapLit($2) }
  /* Graph literal */
  | LBRACE edge_list RBRACE { GraphLit($2) }
  /* Edge literal */
  | edge_lit { $1 }
  /* Array literal */
  | LBRACKET actuals_list RBRACKET { ArrayLit($2) }
  /* Access array/map */
  | ID LBRACKET expr RBRACKET      { Index($1,$3) }
  /* Access struct field */
  | ID PERIOD ID          { StructFieldAccess($1, $3) }

assignable:
    ID                             { Id($1) }
  | ID LBRACKET expr RBRACKET      { Index($1,$3) }
  | ID PERIOD ID                   { StructFieldAccess($1, $3) }

struct_lit:
  LBRACE field_list RBRACE { StructLit($2) }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }

field_list:
    field                      { [$1] }
  | field_list COMMA field     { $3 :: $1 }

field:
  FIELD expr             { ($1, $2) }

kv_list:
    kv                   { [$1] }
  | kv_list COMMA kv     { $3 :: $1 }

kv:
  key_lit COLON expr      { ($1, $3) }

key_lit:
    STRLIT { StrLit($1) }
  | INTLIT { IntLit($1) }
  | CHARLIT { CharLit($1) }
  | struct_lit { $1 }
  | ID      { Id($1) }

/* TODO: (note) disallow non-fully connected graphs */
edge_list:
    edge_lit     { [$1] }
  | edge_list COMMA edge_lit { $3 :: $1 }

edge_lit:
  LPAREN node_lit COMMA node_lit COMMA node_lit RPAREN { EdgeLit($2, $4, $6) }

/* We can probably just replace node_lit with expr to allow any type to be node/edge value? */
node_lit:
    struct_lit  { $1 }
  | ID          { Id($1) }
