(* Semantically-checked Abstract Syntax Tree and functions for printing it *)

open Ast

type sexpr = typ * sx
and sx =
    SIntLit of int
  | SDubLit of float
  | SCharLit of string
  | SStrLit of string
  | SStructLit of (string * sexpr) list
  | SMapLit of (sexpr * sexpr) list
  | SGraphLit of sexpr list
  | SEdgeLit of sexpr * sexpr * sexpr
  | SArrayLit of sexpr list
  | SId of string
  | SArrayIndex of string * sexpr
  | SMapIndex of string * sexpr
  | SStructFieldAccess of string * string
  | SBinop of sexpr * op * sexpr
  | SUnop of uop * sexpr
  | SAssign of sexpr * sexpr
  | SReassign of sexpr * sexpr
  | SCall of string * sexpr list
  | STrue
  | SFalse
  | SNoexpr

type sstmt =
    SBlock of sstmt list
  | SExpr of sexpr
  | SReturn of sexpr
  | SIf of sexpr * sstmt list * sstmt list
  | SIfElseifs of sexpr * sstmt list * (sexpr * sstmt list) list * sstmt list
  | SFor of sexpr * sexpr * sexpr * sstmt list
  | SWhile of sexpr * sstmt list
  | SContinue
  | SBreak

type sfunc_decl = {
    styp : typ;
    sfname : string;
    sformals : bind list;
    slocals : bind list;
    sbody : sstmt list;
  }

type sprogram = bind list * struct_decl list * sfunc_decl list

(* Pretty-printing functions *)

let rec string_of_sexpr (t, e) =
  "(" ^ string_of_typ t ^ " : " ^ (match e with
    SIntLit(l) -> string_of_int l
  | SDubLit(l) -> string_of_float l
  | STrue -> "true"
  | SFalse -> "false"
  | SCharLit(c) -> c
  | SStrLit(s) -> s
  | SArrayLit(l) -> "[" ^ String.concat "," (List.map string_of_sexpr (List.rev l)) ^ "]"
  | SStructLit(l) -> "{" ^ String.concat "," (List.map (fun (f, v) -> f ^ ":" ^ (string_of_sexpr v)) (List.rev l)) ^ "}"
  | SMapLit(l) -> "[" ^ String.concat "," (List.map (fun (k, v) -> (string_of_sexpr k) ^ ":" ^ (string_of_sexpr v)) (List.rev l)) ^ "]"
  | SGraphLit(l) -> "{" ^ String.concat ", " (List.map string_of_sexpr (List.rev l)) ^ "}"
  | SEdgeLit(a, b, c) -> "(" ^ (string_of_sexpr a) ^ ", " ^ (string_of_sexpr b) ^ ", " ^ (string_of_sexpr c) ^ ")"
  | SId(s) -> s
  | SArrayIndex(s, e) | SMapIndex(s, e) -> s ^ "[" ^ (string_of_sexpr e) ^ "]"
  | SStructFieldAccess(s, t) -> s ^ "." ^ t
  | SBinop(e1, o, e2) ->
      string_of_sexpr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr e2
  | SUnop(o, e) -> string_of_uop o ^ string_of_sexpr e
  | SAssign(l, r) -> string_of_sexpr l ^ " := " ^ string_of_sexpr r
  | SReassign(l, r) -> string_of_sexpr l ^ " = " ^ string_of_sexpr r
  | SCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SNoexpr -> ""
				  ) ^ ")"

let rec string_of_sstmt = function
    SBlock(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt (List.rev stmts)) ^ "}\n"
  | SExpr(expr) -> string_of_sexpr expr ^ ";\n";
  | SReturn(expr) -> "return " ^ string_of_sexpr expr ^ ";\n";
  | SIf(e, s, [SBlock([])]) -> "if (" ^ string_of_sexpr e ^ ")\n  " ^
      String.concat "  " (List.map string_of_sstmt (List.rev s))
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr e ^ ")\n  " ^
      String.concat "  " (List.map string_of_sstmt (List.rev s1)) ^
      "else\n  " ^ String.concat "  " (List.map string_of_sstmt (List.rev s2))
  | SIfElseifs(e1, s1, l, [SBlock([])]) -> "if (" ^ string_of_sexpr e1 ^ ")\n  " ^
      String.concat "  " (List.map string_of_sstmt (List.rev s1)) ^ String.concat ""
      (List.map (fun (e2, s2) -> ("elseif (" ^ string_of_sexpr e2 ^ ")\n  " ^ String.concat "  " (List.map string_of_sstmt (List.rev s2)))) l)
  | SIfElseifs(e1, s1, l, s3) -> "if (" ^ string_of_sexpr e1 ^ ")\n  " ^
      String.concat "  " (List.map string_of_sstmt (List.rev s1)) ^ String.concat ""
      (List.map (fun (e2, s2) -> ("elseif (" ^ string_of_sexpr e2 ^ ")\n  " ^ String.concat "  " (List.map string_of_sstmt (List.rev s2)))) l) ^
      "else\n  " ^ String.concat "  " (List.map string_of_sstmt (List.rev s3))
  | SFor(e1, e2, e3, s) ->
      "for (" ^ string_of_sexpr e1  ^ " ; " ^ string_of_sexpr e2 ^ " ; " ^
      string_of_sexpr e3  ^ ") " ^ String.concat "" (List.map string_of_sstmt (List.rev s))
  | SWhile(e, s) -> "while (" ^ string_of_sexpr e ^ ") " ^ String.concat "" (List.map string_of_sstmt (List.rev s))
  | SContinue -> "continue;"
  | SBreak -> "break;"

let string_of_svdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_ssdecl (struct_name, fields) =
  struct_name ^ "{" ^ String.concat "," (List.map (fun (f, t) -> f ^
      ":" ^ (string_of_typ t)) fields)  ^ "}"

let string_of_sfdecl fdecl =
  string_of_typ fdecl.styp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map snd fdecl.sformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.slocals) ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram (vars, structs, funcs) =
  String.concat "" (List.map string_of_svdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_ssdecl structs) ^ "\n" ^
  String.concat "\n" (List.map string_of_sfdecl funcs)
